# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/05_utils.plot.ipynb (unless otherwise specified).

__all__ = ['plot2series', 'add_value_labels', 'plot_yearly_gain']

# Cell
from ..imports import *
from .pandas_hacks import *

# Cell
@delegates(plt.subplots)
def plot2series(s1: Series, s2: Series, names:Optional[str]=None, y_label='y', **kwargs):
    "Plots 2 Series against each other"
    fig, ax = plt.subplots(**kwargs)
    pd.concat([s1, s2], axis=1).plot(ax=ax)
    names = ifnone(names, [s1.name, s2.name])
    ax.legend(names)
    ax.set_ylabel(y_label)
    return ax

# Cell
def add_value_labels(ax, labels=None, spacing=5):
    """Add labels to the end of each bar in a bar chart.

    Arguments:
        ax (matplotlib.axes.Axes): The matplotlib object containing the axes
            of the plot to annotate.
        spacing (int): The distance between the labels and the bars.
    """

    # For each bar: Place a label
    labels = ifnone(labels, [None]*len(ax.patches))
    for rect, label in zip(ax.patches, labels):
        # Get X and Y placement of label from rect.
        y_value = rect.get_height()
        x_value = rect.get_x() + rect.get_width() / 2

        # Number of points between bar and label. Change to your liking.
        space = spacing
        # Vertical alignment for positive values
        va = 'bottom'

        # If value of bar is negative: Place label below bar
        if y_value < 0:
            # Invert space to place label below
            space *= -1
            # Vertically align label at top
            va = 'top'

        # Use Y value as label and format number with one decimal place
        label = "{:.1f}".format(ifnone(label, y_value))

        # Create annotation
        ax.annotate(
            label,                      # Use `label` as label
            (x_value, y_value),         # Place label at end of the bar
            xytext=(0, space),          # Vertically shift label by `space`
            textcoords="offset points", # Interpret `xytext` as offset in points
            ha='center',                # Horizontally center label
            va=va)                      # Vertically align label differently for
                                        # positive and negative values.
    return

# Cell
def plot_yearly_gain(s1: Series, s2: Series, labels=None):
    "Plot bar chart with gain percentage"
    labels = ifnone(labels, ['s1', 's2'])
    _, ax = plt.subplots(figsize=(12,5))
    (pd.concat([s1, s2-s1], axis=1,sort=False)
    .plot.bar(ax=ax, stacked=True, color=['blue', 'violet']))
    ax.set_title('Gain')
    ax.legend(labels)
    gain = 100*(s2/s1-1)
    add_value_labels(ax, gain, spacing=5)
    return


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/07_modelling.mypvfactors.ipynb (unless otherwise specified).

__all__ = ['system_def', 'get_index', 'plot_idx', 'fit_and_plot', 'join_sun', 'get_data', 'fn_build_report',
           'run_pvfactors_simulation', 'individual_report']

# Cell
from ..imports import *
from ..utils.tmy import *
import pvfactors
from pvfactors.geometry import OrderedPVArray
from pvfactors.engine import PVEngine
import pvlib

# Cell
def system_def(albedo: float=0.4,
               n_modules_vertically: int=2,
               module_size: tuple=(1.69,1.01),
               h_ground: float=1,
               surface_tilt: int=38,
               axis_azimuth: int=0,
               surface_azimuth: int=180,
               n_pvrows: int=1,
               tracking: bool=False,
               gcr: float=0.5):
    "A wrapper function to define the pvfactors pvarray_parameters dictionary"
    w_m, h_m = module_size  #width x height of module
    if not tracking: h_center = h_ground + np.sin(2*np.pi*surface_tilt/360)*h_m
    else: h_center = h_ground
    pvarray_parameters = {
        'n_pvrows': n_pvrows,            # number of pv rows
        'pvrow_height': h_center,        # height of pvrows (measured at center / torque tube)
        'pvrow_width': n_modules_vertically * h_m,         # width of pvrows
        'tracking':tracking,
        'axis_azimuth': axis_azimuth,       # azimuth angle of rotation axis
        'surface_tilt': surface_tilt,      # tilt of the pv rows
        'surface_azimuth': surface_azimuth,   # azimuth of the pv rows front surface
        'albedo':albedo,
        'gcr': gcr,               # ground coverage ratio,
        'rho_front_pvrow': 0.075,  # pv row front surface reflectivity
        'rho_back_pvrow': 0.075,    # pv row back surface reflectivity
        'cut':{
            i: {'front': 1,'back': 7} for i in range(n_pvrows) # discretize the front  PV row into 1 segments and back in 7
    }
    }
    return pvarray_parameters

# Cell
def get_index(df, date): return df.index.get_loc(date)

# Cell
def plot_idx(pvarray, idx=4550, with_index=False, figsize=(10,3)):
    "Plot a pvarray at idx"
    f, ax = plt.subplots(figsize=figsize)
    pvarray.plot_at_idx(idx, ax, with_surface_index=with_index)
#     ax.set_xlim(-5, 5)
    return ax

# Cell
def fit_and_plot(data:DataFrame, pvarray_params:dict, date:Union[str, int], with_index:bool=False):
    "Method to debug a pvfactors pvarray, plots the model at idx"
    data_idx = data.loc[date] if type(date) is str else data.iloc[date]
    params = {'axis_azimuth': pvarray_params['axis_azimuth'],
             'gcr': pvarray_params['gcr'],
             'n_pvrows': pvarray_params['n_pvrows'],
             'pvrow_height': pvarray_params['pvrow_height'],
             'pvrow_width': pvarray_params['pvrow_width'],
             'solar_azimuth': data_idx['azimuth'],
             'solar_zenith': data_idx['zenith'],
             'surface_azimuth': pvarray_params['surface_azimuth'],
             'surface_tilt': pvarray_params['surface_tilt'],
             'cut':pvarray_params['cut']}
    pvarray = OrderedPVArray.fit_from_dict_of_scalars(params)
    ax = plot_idx(pvarray, 0, with_index)
    if type(date) is int:
        date = data.index[date]
        date = f'{date.hour}h {date.day} {date.month_name()[0:3]} {date.year}'
    ax.set_title(date)
    return pvarray

# Cell
def join_sun(tmy_data: DataFrame, sunpos:DataFrame, pvarray_params:dict)->DataFrame:
    "Concats meteo data with sunpos"
    data = pd.DataFrame(index=tmy_data.index)
    data['ghi'] = tmy_data.ghi
    data['dni'] = tmy_data.dni
    data['dhi'] = tmy_data.dhi
    data['zenith'] = sunpos.zenith
    data['azimuth'] = sunpos.azimuth
    data['elevation'] = sunpos.elevation
    data['surface_tilt'] = pvarray_params['surface_tilt']
    data['surface_azimuth'] = pvarray_params['surface_azimuth']
    data['albedo'] =  pvarray_params['albedo']
    #doing some patching
    idxs = (data.zenith<90) & (data.ghi<10)
    data.loc[idxs, 'zenith'] = 91.
    return data

# Cell
def get_data(tmy_file:File, pvarray_params:dict, reader:str='pvgis', coerce_year:int=2019):
    "A convenience method to read PVGIS and generic tmy files"
    if reader == 'pvgis':
        gps_data, _, tmy_data = read_pvgis(tmy_file, coerce_year=coerce_year)
    else:
        gps_data, tmy_data = read_tmy(tmy_file, coerce_year=coerce_year)

    sunpos = (pvlib
             .solarposition
             .get_solarposition(tmy_data.index,
                                gps_data['Latitude'],
                                gps_data['Longitude'],
                                gps_data['Elevation'])
             )

    data = join_sun(tmy_data, sunpos, pvarray_params)
    if pvarray_params['tracking']:
        back_track = False
        if pvarray_params['n_pvrows']>1:
            back_track = True
        tracking = (pvlib
                    .tracking
                    .singleaxis(apparent_zenith=data.zenith,
                                apparent_azimuth=data.azimuth,
                                axis_azimuth=pvarray_params['axis_azimuth'],
                                backtrack=back_track,
                                gcr=pvarray_params['gcr'])
                    .fillna(0))
        data.surface_azimuth = tracking.surface_azimuth
        data.surface_tilt = tracking.surface_tilt
    return data

# Cell
fn_build_report=lambda pvarray: pvarray

# Cell
def run_pvfactors_simulation(data: DataFrame, pvarray_params: dict)->OrderedPVArray:
    "Fit and run simulation"
    pvarray = OrderedPVArray.init_from_dict(pvarray_params)
    engine = PVEngine(pvarray)
    pvarray = engine.fit(data.index,
            data.dni,
            data.dhi,
            data.zenith,
            data.azimuth,
            data.surface_tilt,
            data.surface_azimuth,
            data.albedo)
    fn_build_report=lambda pvarray: pvarray
    return engine.run_full_mode(fn_build_report)

# Cell
def individual_report(pvarray:OrderedPVArray, row:int=0, index:pd.Index=None)->DataFrame:
    "Recover individual backface measures"
    back_surface_list = pvarray.ts_pvrows[row].back.list_segments
    index = ifnone(index, range(back_surface_list[0].length.shape[0]))
    irr_dict = {f'qinc_{idx}':s.get_param_weighted('qinc') for idx, s in enumerate(back_surface_list)}
    irr_dict.update({'qinc_front': pvarray.ts_pvrows[row].front.get_param_weighted('qinc'),
                    'qinc_back': pvarray.ts_pvrows[row].back.get_param_weighted('qinc')})
    return pd.DataFrame(irr_dict, index=index)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/06_modelling.mybifacialvf.ipynb (unless otherwise specified).

__all__ = ['CHAMBERY', 'rename_cols', 'format_output_as_pvfactor', 'run_bifacialvf_simulation']

# Cell
from ..imports import *
from ..utils.tmy import read_tmy
import bifacialvf
import pvlib

# Cell
CHAMBERY = {'Name':'Chambery', 'Latitude': 45.637001, 'Longitude': 5.881, 'Elevation': 235.0, 'TZ':-1.0}

# Cell
def rename_cols(df: DataFrame, cols: list=['dni', 'dhi', 'zenith', 'azimuth', 'elevation']):
    "Rename to upper case dni and ghi cols"
    assert set(cols).issubset(df.columns.str.lower()), 'Missing columns on df'
    if not set(['DNI', 'DHI']).issubset(df.columns):
        tmy3 = (df[cols]
                .rename(columns={'dni':'DNI', 'dhi':'DHI'}))
        return tmy3
    else: return df

# Cell
def format_output_as_pvfactor(res: DataFrame, cuts: int):
    "Formats output from bifacialvf as pvfactor's one"
    front_cols = [f'No_{i+1}_RowFrontGTI' for i in range(cuts)]
    back_cols = [f'No_{i+1}_RowBackGTI' for i in range(cuts)]
    aux = pd.DataFrame(index=res.index)
    aux['qinc_front'] = res[front_cols].mean(axis=1)
    aux['qinc_back'] = res[back_cols].mean(axis=1)
    aux[back_cols] = res[back_cols]
    aux =  aux.rename(columns=dict(zip(back_cols,[f'qinc_back_{i}' for i in range(cuts-1, -1, -1)] )))
    return aux[['qinc_front', 'qinc_back']+[f'qinc_back_{i}' for i in range(cuts)]]

# Cell
def run_bifacialvf_simulation(data: DataFrame,
                          pvarray_parameters:dict={'rtr':8.,  'cellRows': 7, 'albedo':0.4},
                          gps_data:dict=CHAMBERY):
    "Run bifacialvf on data, with pvarray_parameters at location"
    outfile = 'output.csv'
    pvarray_parameters.update({'sam_header':False})
    bifacialvf.simulate_inner(rename_cols(data), gps_data, outfile, **pvarray_parameters)
    return (pd.read_csv(outfile, header=2, index_col='date', parse_dates=True)
            .pipe(format_output_as_pvfactor, cuts=pvarray_parameters['cellRows']))


import os

from fastcore.all import *
from fastprogress.fastprogress import progress_bar

import numpy as np
np.seterr(divide='ignore', invalid='ignore')
import pandas as pd
from pandas import DataFrame, Series

pd.set_option('display.max_rows', 999)
pd.set_option('precision', 3)


import hdf5storage

import matplotlib.pyplot as plt

#typing
from typing import Dict, Union, Callable, Optional, Tuple

Number = Union[float, int]
File = Union[Path, str]
All = slice(None)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/03_utils.tmy.ipynb (unless otherwise specified).

__all__ = ['read_pvgis', 'read_tmy']

# Cell
from ..imports import *
import dateutil
import pandas as pd
import pytz

# Cell
def read_pvgis(filename: File, coerce_year: int=None, rename_cols: bool=True):
    "Utility function to read PVGIS tmy files"
    with open(filename, 'r') as csvdata:
        # read in file metadata, advance buffer to second line
        gps_data = {}
        keys = 'Latitude,Longitude,Elevation'.split(',')
        for i in range(3):
            line = csvdata.readline()
            attribute = line.rstrip('\n').split(":")
            gps_data[keys[i]] = float(attribute[1])

        months_year = pd.read_csv(csvdata, nrows=12, engine='python').set_index('month')

        def _parsedate(date_str, year=None):
            "parser to get a coerced year from a tmy file"
            date = pd.datetime.strptime(date_str, '%Y%m%d:%H%M')
            if year is not None:
                date = date.replace(year=year)
            return date


        col_names = 'time(UTC),T2m,RH,G(h),Gb(n),Gd(h),IR(h),WS10m,WD10m,SP'.split(',')
        new_names = 'date,temp,humidity,ghi,dni,dhi,infra,ws,wd,pressure'.split(',')
        date_col = col_names[0]
        data = pd.read_csv(csvdata,
                           header=0,
                           names=col_names,
                           parse_dates=True,
                           date_parser=lambda x: _parsedate(x, year=coerce_year),
                           index_col=date_col,
                           nrows=365*24)
        return gps_data,months_year,data.rename(columns=dict(zip(col_names, new_names)))


# Cell
def read_tmy(filename: File, columns:Optional[list]=None, coerce_year:Optional[int]=None):
    "Reads TMY files with dates on 5 first cols."
    if columns is None:
        columns = ['ghi', 'dni', 'dhi','temp_air', 'temp_dew',
                   'humidity', 'air_pressure','wind_speed', 'wind_dir', 'snow']
    with open(filename, 'r') as csvdata:
        # read in file metadata, advance buffer to second line
        firstline = csvdata.readline().rstrip('\n').split(",")
        secondline = csvdata.readline().rstrip('\n').split(",")
        meta = dict(zip(firstline,secondline))
        meta['Latitude'] = float(meta['Latitude'])
        meta['Longitude'] = float(meta['Longitude'])
        meta['Elevation'] = float(meta['Elevation'])
        meta['Time Zone'] = int(meta['Time Zone'])
        fixed_tz = pytz.FixedOffset(float(meta['Time Zone']) * 60)
        def _parsedate(ymdh, year=None):
            "parser to get a coerced year from a tmy file"
            ymdh = pd.datetime.strptime(ymdh, '%Y %m %d %H %M')
            if year is not None:
                ymdh = ymdh.replace(year=year)
            return ymdh
        data = pd.read_csv(csvdata, header=0,
                           index_col='datetime',
                           parse_dates={'datetime': [0,1,2,3,4]},
                           date_parser=lambda x: _parsedate(x, year=coerce_year)
                          )
        data.columns = columns
        return meta, data.tz_localize(fixed_tz)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/01_utils.pandas_hacks.ipynb (unless otherwise specified).

__all__ = ['group_monthly', 'reindex_monthly_en', 'monthly', 'All', 'mem_usage', 'reduce_mem_usage']

# Cell
from ..imports import *

# Cell
def group_monthly(df: Union[DataFrame, Series], year=2019):
    "Group DataFrame on a monthly basis"
    return (df.groupby([lambda x: x.year, lambda x: x.month]).sum()).loc[year,:]

# Cell
def reindex_monthly_en(data):
    "Rename index to months"
    index=['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun','Jul', 'Aug','Sep', 'Oct', 'Nov', 'Dec']
    if isinstance(data, pd.DataFrame):
        return pd.DataFrame(data.values, index=index, columns=data.columns )
    if isinstance(data, pd.Series):
        return pd.Series(data.values, index=index, name=data.name)

# Cell
def monthly(data, year=2019):
    "Group and reindex a DataFrame or Series monthly"
    return reindex_monthly_en(group_monthly(data, year))

# Cell
def _normalize(df: Union[Series, DataFrame], per_column:bool=False):
    "Scales dataframe to have mean 0 and std 1"
    return (df-df.stack().mean())/df.stack().std() if per_column else (df-df.mean())/df.std()

def _min_max(df: Union[Series, DataFrame], per_column:bool=False):
    "Scales dataframe between 0 and 1"
    return (df-df.stack().min())/(df.stack().max()-df.stack().min()) if per_column else (df-df.min())/(df.max()-df.min())

# Cell
@patch
def normalize(self:DataFrame): return _normalize(self)
@patch
def normalize(self:Series): return _normalize(self)
@patch
def min_max(self:DataFrame): return _min_max(self)
@patch
def min_max(self:Series): return _min_max(self)

# Cell
All = slice(None)

# Cell
def _flatten_cols(df, to_str=True):
    "Flattens the columns MultiIndex"
    df = df.copy()
    flat_cols = df.columns.to_flat_index()
    if to_str:
        df.columns = ['_'.join(c) for c in flat_cols]
    else:
        df.columns = flat_cols
    return df

# Cell
@patch
def flatten_cols(self:DataFrame, to_str=True): return _flatten_cols(self, to_str)

# Cell
def mem_usage(pandas_obj: Union[DataFrame, Series]):
    "Detailed Memory usage of pandas object"
    if isinstance(pandas_obj,pd.DataFrame):
        usage_b = pandas_obj.memory_usage(deep=True).sum()
    else: # we assume if not a df it's a series
        usage_b = pandas_obj.memory_usage(deep=True)
    usage_mb = usage_b / 1024 ** 2 # convert bytes to megabytes
    return f"{usage_mb:03.2f} MB"

def reduce_mem_usage(df: DataFrame, downcast=np.float32, tol:float=1e-7):
    "Downcast float64 to float32 variables on df"
    df_float = df.select_dtypes(include=['float'])
    optimized_df = df.copy()
    optimized_df[df_float.columns] = df_float.astype(dtype=downcast)
    bad_cols = list(df.columns[(optimized_df-df).abs().mean()>tol])
    if len(bad_cols)>0:
        optimized_df.loc[All, bad_cols] = df.loc[All, bad_cols]
        print(f'Columns {bad_cols} where not replaced due to prec. loss')
    return optimized_df


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/03_utils.matlab_io.ipynb (unless otherwise specified).

__all__ = ['read_mat', 'datetime_mtopy', 'get_components', 'dict2tuples', 'make_datetime_index', 'mat2df']

# Cell
from ..imports import *
from .pandas_hacks import reduce_mem_usage
import hdf5storage
import datetime

# Cell
def read_mat(filename: Union[str, Path]):
    mat = hdf5storage.loadmat(str(filename))
    return mat

# Cell
def datetime_mtopy(datenum):
    '''
    Input
        The fractional day count according to datenum datatype in matlab
    Output
        The date and time as a instance of type datetime in python
    Notes on day counting
        matlab: day one is 1 Jan 0000
        python: day one is 1 Jan 0001
        hence a reduction of 366 days, for year 0 AD was a leap year
    EX:
    df['date'] = df['tvec'].apply(lambda tvec: datetime_mtopy(tvec))
    '''
    ii = datetime.datetime.fromordinal(int(datenum) - 366)
    ff = datetime.timedelta(days=datenum%1)
    return ii + ff

def get_components(mat):
    "helper function to process the mat file"
    compList = {}
    for name in mat['dataName']:
        compList[name[0][0][0]]=[]
        for item in name[1]:
            compList[name[0][0][0]]+=[x[0][0] for x in item]
    return compList

def dict2tuples(d):
    "helper to split comp names"
    tuples = []
    for key in d:
        tuples += list(zip(['_'.join(key.split())]*len(d[key]), d[key]))
    return tuples

def make_datetime_index(df):
    "creates a datetime index"
    df['date'] = df[0].apply(lambda tvec: datetime_mtopy(tvec))
    df.set_index('date', inplace=True)
    df.drop(columns=[0], axis=1, inplace=True)
    return

def mat2df(filename):
    'Popdb matlab .mat file to pandas df'
    mat = read_mat(filename)
    data = mat['data']
    tvec = mat['tvec']
    datac = np.concatenate((tvec, data), axis=1)
    df = pd.DataFrame(datac)
    compList = get_components(mat)
    compTuples = dict2tuples(compList)
    make_datetime_index(df)
    df.columns = pd.MultiIndex.from_tuples(compTuples,
                                           names=['component','quantity'])
    return reduce_mem_usage(df)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/04_utils.missing.ipynb (unless otherwise specified).

__all__ = ['is_day', 'plot_missing', 'std_outliers', 'diff_outliers', 'comp', 'view_anomaly', 'stuck_outlier',
           'percentile_cutoff', 'MinMaxRemoval']

# Cell
from ..imports import *
from .pandas_hacks import *
from sklearn.base import TransformerMixin

# Cell
def is_day(df: DataFrame, ghi_col='ghi')->DataFrame:
    "Day test, needs ghi col present"
    assert ghi_col in df.columns, f'{ghi_col} is not present'
    return df[df.loc[All, ghi_col]>0]

# Cell
def plot_missing(df: DataFrame, f:Callable[[DataFrame], DataFrame]=pd.isna,
                 ghi_col:Optional[str]=None):
    "Plot data that satisfies the f  function, optional ghi_col to exclude night"
    if ghi_col is not None:
        df = is_day(df, ghi_col)
    cols = df.columns
    n = len(cols)
    fig, ax = plt.subplots(figsize=(15,n*0.3))
    outliers = f(df).astype(int)
    cax = ax.matshow(outliers.T, interpolation=None, aspect='auto')
    #Color Bar
    cb = fig.colorbar(cax)
    cb.set_ticks(range(outliers.max().max()+1))
    cb.set_label('outlier values')
    #Matrix plot
    xticks = np.arange(0, len(outliers), int(len(outliers)/7))
    ax.set_xticks(xticks)
    ax.set_xticklabels([outliers.index[i].strftime('%d-%m-%Y') for i in xticks])
    ax.set_yticks(range(n))
    ytickslabels = [str(name)+f' ({100*outliers[name].sum()/len(outliers[name]):.3f}%)' for name in cols.values]
    ax.set_yticklabels(ytickslabels)
    if legend is not None: ax.legend(legend)
    return

# Cell
def std_outliers(mean=0, std=1, std_coef=3):
    "Std outlier test func"
    def _inner(df): return (np.abs(df-mean)>std_coef*std)
    return _inner

def diff_outliers(std=1, std_coef=2):
    "Std test on 1st derivative"
    def _inner(df): return (df.diff().abs()>std_coef*std)
    return _inner

# Cell
def comp(l:list, numbered=True):
    "Compose a list of outlier functions, return different values if `numbered=True`"
    def _inner(df):
        n = len(l)
        if numbered:
            return reduce(operator.add, [(i+1)*f(df).astype(int) for i, f in enumerate(l)])
        else:
            return reduce(operator.add, [f(df).astype(int) for f in l])
    return _inner

# Cell
def view_anomaly(s: Series, f=std_outliers(0,1,3)):
    "Plots the missing/anomaly points"
    _, ax = plt.subplots(figsize=(14,6))
    a = f(s)
    ax.set_title(f' Anomaly rate: ({100*a.values.sum()/len(s):.3f}%)')
    ax.plot(s, color='blue', label = 'Normal')
    ax.scatter(a[a].index, s[a].fillna(0), color='red', label = 'Anomaly')
    ax.legend()
    return ax

# Cell
def stuck_outlier(tol:float=1e-3, periods=1):
    "Detects if the values are stuck"
    def _inner(df):
        return df.diff().abs().rolling(periods, center=True).max()<tol
    return _inner

# Cell
def percentile_cutoff(data, p_low=5, p_high=95):
    q_low = np.percentile(data,p_low)
    q_high = np.percentile(data, p_high)
    iqr = q_high - q_low
    cut_off = iqr * 1.25
    lower, upper = q_low - cut_off, q_high + cut_off
    return lower, upper

# Cell
class MinMaxRemoval(TransformerMixin):
    "Remove values outside Min-Max within a given tolerance"
    def __init__(self, min_max_tol_dict:dict):
        "A dictionary with the [min, max] interval and a tolerance"
        self.d = min_max_tol_dict

    def fit(self, X, y=None):
        return self

    def transform(self, X:DataFrame, y=None):
        "Apply the transform to your data df"

        X = X.copy()
        for c, val in self.d.items():
            m, M, tol = val
            X[c] = (X[c].mask(np.abs(X[c] - m) < tol, m+tol)  #min replace
                        .mask(np.abs(X[c] - M) < tol, M-tol)  #max replace
                        .where((m - tol < X[c]) & (X[c] < M + tol), np.nan)  #else replace by NaN
                   )
        return X


# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"get_files": "00_utils.core.ipynb",
         "group_monthly": "01_utils.pandas_hacks.ipynb",
         "reindex_monthly_en": "01_utils.pandas_hacks.ipynb",
         "monthly": "01_utils.pandas_hacks.ipynb",
         "DataFrame.normalize": "01_utils.pandas_hacks.ipynb",
         "Series.normalize": "01_utils.pandas_hacks.ipynb",
         "DataFrame.min_max": "01_utils.pandas_hacks.ipynb",
         "Series.min_max": "01_utils.pandas_hacks.ipynb",
         "All": "01_utils.pandas_hacks.ipynb",
         "DataFrame.flatten_cols": "01_utils.pandas_hacks.ipynb",
         "mem_usage": "01_utils.pandas_hacks.ipynb",
         "reduce_mem_usage": "01_utils.pandas_hacks.ipynb",
         "read_mat": "03_utils.matlab_io.ipynb",
         "datetime_mtopy": "03_utils.matlab_io.ipynb",
         "get_components": "03_utils.matlab_io.ipynb",
         "dict2tuples": "03_utils.matlab_io.ipynb",
         "make_datetime_index": "03_utils.matlab_io.ipynb",
         "mat2df": "03_utils.matlab_io.ipynb",
         "read_pvgis": "03_utils.tmy.ipynb",
         "read_tmy": "03_utils.tmy.ipynb",
         "is_day": "04_utils.missing.ipynb",
         "plot_missing": "04_utils.missing.ipynb",
         "std_outliers": "04_utils.missing.ipynb",
         "diff_outliers": "04_utils.missing.ipynb",
         "comp": "04_utils.missing.ipynb",
         "view_anomaly": "04_utils.missing.ipynb",
         "stuck_outlier": "04_utils.missing.ipynb",
         "percentile_cutoff": "04_utils.missing.ipynb",
         "MinMaxRemoval": "04_utils.missing.ipynb",
         "plot2series": "05_utils.plot.ipynb",
         "add_value_labels": "05_utils.plot.ipynb",
         "plot_yearly_gain": "05_utils.plot.ipynb",
         "CHAMBERY": "06_modelling.mybifacialvf.ipynb",
         "rename_cols": "06_modelling.mybifacialvf.ipynb",
         "format_output_as_pvfactor": "06_modelling.mybifacialvf.ipynb",
         "run_bifacialvf_simulation": "06_modelling.mybifacialvf.ipynb",
         "system_def": "07_modelling.mypvfactors.ipynb",
         "get_index": "07_modelling.mypvfactors.ipynb",
         "plot_idx": "07_modelling.mypvfactors.ipynb",
         "fit_and_plot": "07_modelling.mypvfactors.ipynb",
         "join_sun": "07_modelling.mypvfactors.ipynb",
         "get_data": "07_modelling.mypvfactors.ipynb",
         "fn_build_report": "07_modelling.mypvfactors.ipynb",
         "run_pvfactors_simulation": "07_modelling.mypvfactors.ipynb",
         "individual_report": "07_modelling.mypvfactors.ipynb"}

modules = ["utils/core.py",
           "utils/pandas_hacks.py",
           "utils/matlab_io.py",
           "utils/tmy.py",
           "utils/missing.py",
           "utils/plot.py",
           "modelling/mybifacialvf.py",
           "modelling/mypvfactors.py"]

doc_url = "https://tcapelle.github.io/pvtools/"

git_url = "https://github.com/tcapelle/pvtools/tree/master/"

def custom_doc_links(name): return None



# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/00_utils.core.ipynb (unless otherwise specified).

__all__ = ['get_files']

# Cell
from ..imports import *

# Cell
def _get_files(p, fs, extensions=None):
    p = Path(p)
    res = [p/f for f in fs if not f.startswith('.')
           and ((not extensions) or f'.{f.split(".")[-1].lower()}' in extensions)]
    return res

# Cell
def get_files(path, extensions=None, recurse=True, folders=None):
    "Get all the files in `path` with optional `extensions`, optionally with `recurse`, only in `folders`, if specified."
    path = Path(path)
    folders=L(folders)
    extensions = setify(extensions)
    extensions = {e.lower() for e in extensions}
    if recurse:
        res = []
        for i,(p,d,f) in enumerate(os.walk(path)): # returns (dirpath, dirnames, filenames)
            if len(folders) !=0 and i==0: d[:] = [o for o in d if o in folders]
            else:                         d[:] = [o for o in d if not o.startswith('.')]
            res += _get_files(p, f, extensions)
    else:
        f = [o.name for o in os.scandir(path) if o.is_file()]
        res = _get_files(path, f, extensions)
    return res